# Prochází podstrom složek a souborů, simuluje tedy linuxový příkaz find.
# Narazí-li na složku, do které nemá právo vstoupit nebo ji číst, nebude se o to ani pokoušet.
# Copyright © 2009 Dan Zeman <zeman@ufal.mff.cuni.cz>
# Licence: GNU GPL

package find;
use utf8;



#------------------------------------------------------------------------------
# Projde podstrom složek a souborů. Na každou složku a soubor s výjimkou '.' a
# '..' zavolá callback funkci, která může se jménem souboru udělat, co chce, a
# u složky navíc musí vrátit 1, pokud se do složky má vlézt dovnitř.
#
# Ukázkové volání: find::go('.', \&find::print);
#------------------------------------------------------------------------------
sub go
{
    my $cesta = shift;
    my $callback = shift;
    # Vede cesta ke složce, do které máme právo vstoupit a číst její obsah?
    unless(-d $cesta && -r $cesta && -x $cesta)
    {
        return 0;
    }
    # Otevřít složku a načíst její obsah.
    # Pokud to navzdory výše provedenému testu nejde, je něco špatně a nezbývá než hodit výjimku.
    opendir(DIR, $cesta) or die("Nelze číst obsah složky $cesta: $!\n");
    my @obsah = readdir(DIR);
    closedir(DIR);
    # Projít obsah složky a roztřídit si ho podle druhu.
    my %druh;
    foreach my $objekt (@obsah)
    {
        # Odkaz na aktuální složku a jejího rodiče zcela ignorovat.
        if($objekt =~ m/^\.\.?$/)
        {
            $druh{$objekt} = '.';
        }
        # U složek nás zajímá, zda je smíme procházet a číst.
        elsif(-d "$cesta/$objekt")
        {
            if(-r "$cesta/$objekt" && -x "$cesta/$objekt")
            {
                $druh{$objekt} = 'drx';
            }
            else
            {
                $druh{$objekt} = 'd';
            }
        }
        # Soubor nebo cokoli jiného (symbolický odkaz?)
        else
        {
            $druh{$objekt} = 'o';
        }
    }
    # Projít obsah složky ještě jednou. Tentokrát volat callback funkci a případně rovnou procházet děti.
    foreach my $objekt (@obsah)
    {
        next if($druh{$objekt} eq '.');
        my $vysledek = &{$callback}($cesta, $objekt, $druh{$objekt});
        if($druh{$objekt} eq 'drx' && $vysledek)
        {
            go("$cesta/$objekt", $callback);
        }
    }
}



#------------------------------------------------------------------------------
# Ukázková callback funkce, která vytiskne úplnou cestu k objektu na standardní
# výstup.
#------------------------------------------------------------------------------
sub print
{
    my $cesta = shift;
    my $objekt = shift;
    my $druh = shift;
    print("$cesta/$objekt\n");
    return $druh eq 'drx';
}



1;
